from abc import ABC, abstractmethod


class MetadataSource(ABC):
    """Metadata Source contract."""

    @property
    @abstractmethod
    def type(self) -> str:
        """
        The type of the source.
        :return: the name o of the source.
        """
        pass



from dataclasses import dataclass, field
from typing import List, NamedTuple, Tuple

from loguru import logger
from rich.console import Console
from rich.table import Table

from .data_rules import MetadataGuardianResults


@dataclass
class ReportResults:
    """Metadata Guardian Results."""

    source: str
    results: List[MetadataGuardianResults] = field(default_factory=list)


@dataclass
class MetadataGuardianReport:
    """Metadata Guardian Report."""

    report_results: List[ReportResults] = field(default_factory=list)

    def append(self, other_report: "MetadataGuardianReport") -> None:
        """
        Concat the results before making the report.
        :param other_report: other report to append
        :return:
        """
        self.report_results = self.report_results + other_report.report_results

    def to_console(self) -> None:
        """
        Display the metadata guardian results to the console.
        :return:
        """
        _console = Console()
        _table = Table(
            title="MetadataGuardian report", show_header=True, header_style="bold dim"
        )
        _table.add_column("Category", style="yellow", width=30)
        _table.add_column("Source", style="cyan", width=30)
        _table.add_column("Content", style="cyan", width=30)
        _table.add_column("Name", style="magenta", width=30)
        _table.add_column("Pattern")
        _table.add_column("Documentation", width=50)
        for report in self.report_results:
            for result in report.results:
                for data_rule in result.data_rules:
                    _table.add_row(
                        result.category,
                        report.source,
                        result.content,
                        data_rule.rule_name,
                        data_rule.regex_pattern,
                        data_rule.documentation,
                    )
        if _table.rows:
            logger.warning(f"MetadataGuardian detected data rules violation(s).")
            _console.print(_table)
        else:
            logger.info(
                f"No data rules violation(s) were detected by MetadataGuardian."
            )



import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text, Union

from .local_metadata_source import LocalMetadataSource


@dataclass
class AvroSchemaSource(LocalMetadataSource):
    """Instance for a local Avro Schema file."""

    def read(self) -> Union[Text, bytes]:
        """Read the AVRO Schema file."""
        with open(self.local_path, "r") as file:
            return file.read()

    def schema(self) -> Dict[str, Any]:
        """
        Get the AVRO schema.
        :return: the schema
        """
        content = self.read()
        return json.loads(content)

    def get_field_attribute(self, attribute_name: str) -> Optional[List[str]]:
        """
        Get the specific attribute from the AVRO Schema file.
        :param attribute_name: the attribute name to get
        :return: the list of attributes in the fields
        """
        return [
            field[attribute_name] if attribute_name in field else None
            for field in self.schema()["fields"]
        ]

    def get_column_names(self) -> List[str]:
        """
        Get column names from the AVRO Schema file.
        :return: the list of the column names
        """
        return [field["name"] for field in self.schema()["fields"]]

    @property
    def namespace(self) -> str:
        """
        Namespace of the AVRO schema.
        :return: the namespace
        """
        return self.schema()["namespace"]

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name o of the source.
        """
        return "LocalAvroSchema"



from metadata_guardian.data_rules import DataRules, MetadataGuardianResults
from metadata_guardian.report import MetadataGuardianReport, ReportResults


def test_report_append_results():
    category = "category"
    content = "content"
    second_category = "second_category"
    second_content = "second_content"
    results = ReportResults(
        source="source1",
        results=[
            MetadataGuardianResults(category=category, content=content, data_rules=[])
        ],
    )
    other_report = MetadataGuardianReport(
        report_results=[
            ReportResults(
                source="source2",
                results=[
                    MetadataGuardianResults(
                        category=second_category, content=second_content, data_rules=[]
                    )
                ],
            )
        ]
    )

    report = MetadataGuardianReport(report_results=[results])
    report.append(other_report)

    assert report.report_results[0].source == "source1"
    assert report.report_results[1].source == "source2"
    assert report.report_results[0].results[0].category == category
    assert report.report_results[1].results[0].category == second_category
    assert report.report_results[0].results[0].content == content
    assert report.report_results[1].results[0].content == second_content



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import pyarrow
from pyarrow.orc import ORCFile

from .local_metadata_source import LocalMetadataSource


@dataclass
class ORCSource(LocalMetadataSource):
    """Instance for a local ORC file."""

    def read(self) -> ORCFile:
        """
        Read the ORC file.
        :return:
        """
        return ORCFile(self.local_path)

    def schema(self) -> pyarrow.Schema:
        """
        Get the ORC File.
        :return: the orc schema
        """
        return self.read().schema

    def get_column_names(self) -> List[str]:
        """
        Get the column names from the schema.
        :return: the list of the column names
        """
        return self.schema().names

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name of the source.
        """
        return "LocalORC"



import pytest

from metadata_guardian.source.local.orc_source import ORCSource


@pytest.mark.parametrize("local_file", ["example.orc"], indirect=["local_file"])
def test_orc_source(local_file):
    source = ORCSource(local_path=local_file)
    expected = [
        "boolean1",
        "byte1",
        "short1",
        "int1",
        "long1",
        "float1",
        "double1",
        "bytes1",
        "string1",
        "middle",
        "list",
        "map",
    ]

    column_names = source.get_column_names()

    assert expected == column_names



import importlib.resources
from dataclasses import dataclass
from enum import Enum
from typing import List

from .metadata_guardian import RawDataRules


class AvailableCategory(Enum):
    """Available Data Rules Categories."""

    PII = "pii_rules.yaml"
    INCLUSION = "inclusion_rules.yaml"


@dataclass
class DataRule:
    """DataRule instance with a regex pattern and a documentation."""

    rule_name: str
    regex_pattern: str
    documentation: str


@dataclass
class MetadataGuardianResults:
    """Metadata Guardian Results instance with the content that matches with the data rules."""

    category: str
    content: str
    data_rules: List[DataRule]


@dataclass(init=False)
class DataRules:
    """Data Rules instances."""

    def __init__(self, path: str) -> None:
        self._data_rules = RawDataRules(path)

    @classmethod
    def from_available_category(cls, category: AvailableCategory) -> "DataRules":
        """
        Get Data Rules from an available category.
        :param category: the available category of the data rules
        :return: the Data Rules instance
        """
        if not isinstance(category, AvailableCategory):
            raise ValueError("The category must be an instance of AvailableCategory")
        with importlib.resources.path(
            "metadata_guardian.rules", category.value
        ) as resource:
            path = str(resource)
            return cls(path=path)

    def validate_word(self, word: str) -> MetadataGuardianResults:
        """
        Validate a word with the data rules defined.
        :param word: the word to validate
        :return: the metadata guardian results
        """
        result = self._data_rules.validate_word(word=word)
        return MetadataGuardianResults(
            category=result._category,
            content=result._content,
            data_rules=[
                DataRule(
                    data_rule.rule_name,
                    data_rule.pattern,
                    data_rule.documentation,
                )
                for data_rule in result._data_rules
            ],
        )

    def validate_words(self, words: List[str]) -> List[MetadataGuardianResults]:
        """
        Validate a list of words with the data rules defined.:param word: the word to validate
        :param words: the words to validate
        :return: the metadata guardian results
        """
        results = self._data_rules.validate_words(words=words)
        return [
            MetadataGuardianResults(
                category=result._category,
                content=result._content,
                data_rules=[
                    DataRule(
                        data_rule.rule_name,
                        data_rule.pattern,
                        data_rule.documentation,
                    )
                    for data_rule in result._data_rules
                ],
            )
            for result in results
        ]

    def validate_file(self, path: str) -> List[MetadataGuardianResults]:
        """
        Validate a file content with the data rules defined.
        :param path: the file path
        :return: the metadata guardian results
        """
        results = self._data_rules.validate_file(path)
        return [
            MetadataGuardianResults(
                category=result._category,
                content=result._content,
                data_rules=[
                    DataRule(
                        data_rule.rule_name,
                        data_rule.pattern,
                        data_rule.documentation,
                    )
                    for data_rule in result._data_rules
                ],
            )
            for result in results
        ]

    def __str__(self) -> str:
        return f"DataRules({[str(data_rule) for data_rule in self._data_rules]})"

    def __repr__(self) -> str:
        return f"DataRules({[str(data_rule) for data_rule in self._data_rules]})"



from dataclasses import dataclass
from typing import Any, List, Optional

import boto3
import botocore
from loguru import logger
from mypy_boto3_athena.client import AthenaClient
from mypy_boto3_glue.client import GlueClient

from .external_metadata_source import ExternalMetadataSource


@dataclass
class AthenaSource(ExternalMetadataSource):
    """Athena Source instance."""

    s3_staging_dir: str
    catalog_name: str = "AWSDataCatalog"
    region_name: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None

    def get_connection(self) -> AthenaClient:
        """
        Get Athena connection.
        :return: a new Athena connection.
        """
        return boto3.client(
            "athena",
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )

    def get_column_names(
        self, database_name: str, table_name: str, include_comment: bool = False
    ) -> List[str]:
        """
        Get column names from the table.
        :param database_name: the database name
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """
        try:
            client = self.get_connection()
            response = client.get_table_metadata(
                CatalogName=self.catalog_name,
                DatabaseName=database_name,
                TableName=table_name,
            )
            columns = list()
            for row in response["TableMetadata"]["Columns"]:
                columns.append(row["Name"].lower())
                if include_comment:
                    columns.append(row["Comment"].lower())
            return columns
        except botocore.exceptions.ClientError as error:
            logger.exception(
                f"Error in getting columns name from AWS Athena {database_name}.{table_name} for catalog {self.catalog_name}"
            )
            raise error

    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the database in AWS Athena.
        :param database_name: the database name
        :return: the list of the table names of the database
        """
        try:
            client = self.get_connection()
            table_names_list = list()
            response = client.list_table_metadata(
                CatalogName=self.catalog_name,
                DatabaseName=database_name,
            )
            for table in response["TableMetadataList"]:
                table_names_list.append(table["Name"])
            while "NextToken" in response:
                response = client.list_table_metadata(
                    CatalogName=self.catalog_name,
                    DatabaseName=database_name,
                    NextToken=response["NextToken"],
                )
                for table in response["TableMetadataList"]:
                    table_names_list.append(table["Name"])
            return table_names_list
        except botocore.exceptions.ClientError as error:
            logger.exception(
                f"Error in getting table names list from AWS Athena from the database {database_name} for catalog {self.catalog_name}"
            )
            raise error

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name o of the source.
        """
        return "AWSAthena"


@dataclass
class GlueSource(ExternalMetadataSource):
    """Glue Source instance."""

    region_name: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None

    def get_connection(self) -> GlueClient:
        """
        Get the Glue connection
        :return: one Glue client
        """
        return boto3.client(
            "glue",
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )

    def get_column_names(
        self, database_name: str, table_name: str, include_comment: bool = False
    ) -> List[str]:
        """
        Get the column names from AWS Glue table.
        :param database_name: the name of the database
        :param table_name: the name of the table
        :param include_comment: include the comments
        :return: the list of the column names
        """
        try:
            client = self.get_connection()
            response = client.get_table(DatabaseName=database_name, Name=table_name)
            columns = list()
            for row in response["Table"]["StorageDescriptor"]["Columns"]:
                columns.append(row["Name"].lower())
                if include_comment:
                    columns.append(row["Comment"].lower())
            return columns
        except botocore.exceptions.ClientError as error:
            logger.exception(
                f"Error in getting columns name from AWS Glue from the table {database_name}.{table_name}"
            )
            raise error

    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the database in AWS Glue.
        :param database_name: the database name
        :return: the list of the table names of the database
        """
        try:
            client = self.get_connection()
            table_names_list = list()
            response = client.get_tables(
                DatabaseName=database_name,
            )
            for table in response["TableList"]:
                table_names_list.append(table["Name"])
            while "NextToken" in response:
                response = client.list_table_metadata(
                    DatabaseName=database_name, NextToken=response["NextToken"]
                )
                for table in response["TableList"]:
                    table_names_list.append(table["Name"])
            return table_names_list
        except botocore.exceptions.ClientError as error:
            logger.exception(
                f"Error in getting table names list from AWS Glue from the database {database_name}"
            )
            raise error

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name of the source.
        """
        return "AWSGlue"



import asyncio
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from loguru import logger
from pyarrow import cpu_count

from .data_rules import DataRules
from .report import MetadataGuardianReport, ReportResults
from .source import ExternalMetadataSource, LocalMetadataSource, MetadataSource


class Scanner(ABC):
    """Scanner interface."""

    @abstractmethod
    def scan_local(self, source: LocalMetadataSource) -> MetadataGuardianReport:
        """
        Scan the column names from the local source.
        :param source: the LocalMetadataSource to scan
        :return: a Metadata Guardian report
        """
        pass

    @abstractmethod
    def scan_external(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source.
        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        pass

    @abstractmethod
    async def scan_external_async(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        tasks_limit: int,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source asynchronously.
        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param tasks_limit: the limit of the tasks to run in parallel
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        pass


@dataclass
class ColumnScanner(Scanner):
    """Column Scanner instance."""

    data_rules: DataRules

    def scan_local(self, source: LocalMetadataSource) -> MetadataGuardianReport:
        """
        Scan the column names from the local source.
        :param source: the MetadataSource to scan
        :return: a Metadata Guardian report
        """
        return MetadataGuardianReport(
            report_results=[
                ReportResults(
                    source=source.local_path,
                    results=self.data_rules.validate_words(
                        words=source.get_column_names()
                    ),
                )
            ]
        )

    def scan_external(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source using a table name or a database name.
        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        if table_name:
            logger.debug(
                f"Get the column names list from the table {database_name}.{table_name}"
            )
            report = MetadataGuardianReport(
                report_results=[
                    ReportResults(
                        source=f"{database_name}.{table_name}",
                        results=self.data_rules.validate_words(
                            words=source.get_column_names(
                                database_name=database_name,
                                table_name=table_name,
                                include_comment=include_comment,
                            )
                        ),
                    )
                ]
            )
        else:
            report = MetadataGuardianReport()
            logger.debug(
                f"Get the table names list from the database {database_name} for {source.type}"
            )
            table_names_list = source.get_table_names_list(database_name=database_name)
            for table_name in table_names_list:
                logger.debug(
                    f"Get the column names list from the table {database_name}.{table_name}"
                )
                report.append(
                    MetadataGuardianReport(
                        report_results=[
                            ReportResults(
                                source=f"{database_name}.{table_name}",
                                results=self.data_rules.validate_words(
                                    words=source.get_column_names(
                                        database_name=database_name,
                                        table_name=table_name,
                                        include_comment=include_comment,
                                    )
                                ),
                            )
                        ]
                    )
                )
        return report

    async def scan_external_async(
        self,
        source: ExternalMetadataSource,
        database_name: str,
        tasks_limit: int = cpu_count(),
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> MetadataGuardianReport:
        """
        Scan the column names from the external source using a table name or a database name.
        Note that it can generate multiple concurrent calls to your metadata source.

        :param source: the ExternalMetadataSource to scan
        :param database_name: the name of the database
        :param tasks_limit: the limit of the tasks to run in parallel
        :param table_name: the name of the table
        :param include_comment: the scan include the comment section
        :return: a Metadata Guardian report
        """
        semaphore = asyncio.Semaphore(tasks_limit)

        async def async_validate_words(table_name: str) -> ReportResults:
            async with semaphore:
                logger.debug(
                    f"Get the column names list from the table {database_name}.{table_name}"
                )
                loop = asyncio.get_event_loop()
                words = await loop.run_in_executor(
                    None,
                    source.get_column_names,
                    database_name,
                    table_name,
                    include_comment,
                )
                return ReportResults(
                    source=f"{database_name}.{table_name}",
                    results=self.data_rules.validate_words(words=words),
                )

        if table_name:
            tasks = [async_validate_words(table_name=table_name)]
        else:
            table_names_list = source.get_table_names_list(database_name=database_name)
            logger.debug(
                f"Get the table names list from the database {database_name} for {source.type}"
            )

            tasks = [
                async_validate_words(table_name=table_name)
                for table_name in table_names_list
            ]
        report_results = await asyncio.gather(*tasks)
        report = MetadataGuardianReport(report_results=report_results)
        return report


@dataclass
class ContentFilesScanner:
    """Content Files Scanner instance."""

    data_rules: DataRules

    def scan_local_file(self, path: str) -> MetadataGuardianReport:
        """
        Scan a file with data rules.
        :param path: the path of the file to scan
        :return: a Metadata Guardian report
        """
        return MetadataGuardianReport(
            report_results=[
                ReportResults(
                    source=path, results=self.data_rules.validate_file(path=path)
                )
            ]
        )

    def scan_directory(
        self, directory_path: str, file_names_extension: str
    ) -> MetadataGuardianReport:
        """
        Scan all the files inside directory path with the file name extension.
        :param directory_path: the directory path to scan
        :param file_names_extension: the file name extension to include (without the .)
        :return: a Metadata Guardian report
        """
        report = MetadataGuardianReport()
        for root, dirs, files in os.walk(directory_path):
            for name in files:
                if name.endswith(f".{file_names_extension}"):
                    report.append(
                        other_report=self.scan_local_file(path=f"{root}/{name}")
                    )
        return report



from dataclasses import dataclass

from .local_metadata_source import LocalMetadataSource


@dataclass
class ParquetSource(LocalMetadataSource):
    """Instance for a local Parquet file."""

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name of the source.
        """
        return "LocalParquet"



from unittest.mock import patch

from deltalake import Field, Schema

from metadata_guardian.source.external.deltatable_source import DeltaTableSource


@patch(
    "metadata_guardian.source.external.deltatable_source.DeltaTableSource.get_connection"
)
def test_deltatable_source_get_column_names(mock_connection):
    uri = "s3://test_table"
    schema = Schema(
        fields=[
            Field(
                name="timestamp",
                type="string",
                nullable=False,
                metadata={"comment": "comment1"},
            ),
            Field(
                name="address_id",
                type="int",
                nullable=False,
                metadata={"comment": "comment2"},
            ),
        ],
        json_value={},
    )
    mock_connection.return_value = mock_connection
    mock_connection.schema.return_value = schema
    expected = [
        "timestamp",
        "{'comment': 'comment1'}",
        "address_id",
        "{'comment': 'comment2'}",
    ]

    column_names = DeltaTableSource(uri=uri).get_column_names(include_comment=True)

    assert column_names == expected


@patch("deltalake.DeltaTable.from_data_catalog")
def test_deltatable_source_get_column_names_from_database_and_table(mock_connection):
    uri = "s3://test_table"
    database_name = "database_name"
    table_name = "table_name"
    schema = Schema(
        fields=[
            Field(
                name="timestamp",
                type="string",
                nullable=False,
                metadata={"comment": "comment1"},
            ),
            Field(
                name="address_id",
                type="int",
                nullable=False,
                metadata={"comment": "comment2"},
            ),
        ],
        json_value={},
    )
    mock_connection.return_value = mock_connection
    mock_connection.schema.return_value = schema
    expected = [
        "timestamp",
        "{'comment': 'comment1'}",
        "address_id",
        "{'comment': 'comment2'}",
    ]

    column_names = DeltaTableSource(uri=uri).get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert column_names == expected



from abc import abstractmethod
from typing import Any, List

from ..metadata_source import MetadataSource


class ExternalMetadataSource(MetadataSource):
    """ExternalMetadataSource Source."""

    @abstractmethod
    def get_column_names(
        self, database_name: str, table_name: str, include_comment: bool = False
    ) -> List[str]:
        """
        Get the column names from the schema.
        :param database_name: the database name
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """
        pass

    @abstractmethod
    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the database.
        :param database_name: the database name
        :return: the list of the table names of the database
        """
        pass

    @abstractmethod
    def get_connection(self) -> Any:
        """
        Get the connection of the source.
        :return: the source connection
        """
        pass



import pytest

from metadata_guardian.source.local.parquet_source import ParquetSource


@pytest.mark.parametrize(
    "local_file", ["subscriptions.parquet"], indirect=["local_file"]
)
def test_parquet_source_column_names(local_file):
    source = ParquetSource(local_path=local_file)
    expected = ["timestamp", "id", "email"]

    column_names = source.get_column_names()

    assert expected == column_names



from dataclasses import dataclass
from typing import Any, List, Optional

from deltalake import DataCatalog, DeltaTable
from loguru import logger

from .external_metadata_source import ExternalMetadataSource


@dataclass
class DeltaTableSource(ExternalMetadataSource):
    uri: str
    data_catalog: DataCatalog = DataCatalog.AWS

    def get_connection(self) -> Any:
        """
        Get the DeltaTable instance.
        :return: the DeltaTable instance
        """
        return DeltaTable(self.uri)

    def get_column_names(
        self,
        database_name: Optional[str] = None,
        table_name: Optional[str] = None,
        include_comment: bool = False,
    ) -> List[str]:
        """
        Get column names from the Delta table.
        :param database_name: the database name
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """
        try:
            if database_name and table_name:
                delta_table = DeltaTable.from_data_catalog(
                    data_catalog=self.data_catalog,
                    database_name=database_name,
                    table_name=table_name,
                )
            else:
                delta_table = self.get_connection()
            schema = delta_table.schema()
            columns = list()
            for field in schema.fields:
                columns.append(field.name.lower())
                if include_comment and field.metadata:
                    columns.append(str(field.metadata).lower())
            return columns
        except Exception as error:
            logger.exception(
                f"Error in getting columns name from the DeltaTable {self.uri}"
            )
            raise error

    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the database.
        :param database_name: the database name
        :return: the list of the table names of the database
        """
        raise NotImplemented()

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name o of the source.
        """
        return "DeltaTable"



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from google.cloud import bigquery
from loguru import logger

from .external_metadata_source import ExternalMetadataSource


@dataclass
class BigQuerySource(ExternalMetadataSource):
    """Instance of a BigQuery source."""

    service_account_json_path: str
    project: Optional[str] = None
    location: Optional[str] = None

    def get_connection(self) -> Any:
        """
        Get the Big Query connection.
        :return: a BigQuery Client
        """
        try:
            return bigquery.Client.from_service_account_json(
                self.service_account_json_path,
                project=self.project,
                location=self.location,
            )
        except Exception as exception:
            logger.exception("Error when connecting to BigQuery")
            raise exception

    def get_column_names(
        self, database_name: str, table_name: str, include_comment: bool = False
    ) -> List[str]:
        """
        Get column names from the table.
        :param database_name: in that case the dataset
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """

        try:
            client = self.get_connection()
            query_job = client.query(
                f'SELECT column_name, description FROM `{database_name}.INFORMATION_SCHEMA.COLUMN_FIELD_PATHS` WHERE table_name = "{table_name}"'
            )
            results = query_job.result()
            columns = list()
            for row in results:
                columns.append(row.column_name.lower())
                if include_comment:
                    columns.append(row.description.lower())
            return columns
        except Exception as exception:
            logger.exception(
                f"Error in getting columns name from BigQuery {database_name}.{table_name}"
            )
            raise exception

    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the GCP dataset.
        :param database_name: in that case the dataset
        :return: the list of the table names list
        """

        try:
            client = self.get_connection()
            query_job = client.query(
                f"SELECT table_name FROM `{database_name}.INFORMATION_SCHEMA.TABLES`"
            )
            results = query_job.result()
            table_names_list = list()
            for row in results:
                table_names_list.append(row.table_name.lower())
            return table_names_list
        except Exception as exception:
            logger.exception(
                f"Error in getting the table names list name from BigQuery {database_name}"
            )
            raise exception

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name bof the source.
        """
        return "BigQuery"



from unittest.mock import patch

from metadata_guardian.source.external.aws_source import AthenaSource, GlueSource


@patch("boto3.client")
def test_athena_source_get_column_names(mock_connection):
    database_name = "test_database"
    table_name = "test_table"
    s3_staging_dir = "s3_test"
    response = {
        "TableMetadata": {
            "Name": table_name,
            "TableType": "EXTERNAL_TABLE",
            "Columns": [
                {"Name": "timestamp", "Type": "string", "Comment": "comment1"},
                {"Name": "address_id", "Type": "int", "Comment": "comment2"},
            ],
            "PartitionKeys": [],
            "Parameters": {},
        }
    }
    mock_connection.return_value = mock_connection
    mock_connection.get_table_metadata.return_value = response
    expected = ["timestamp", "comment1", "address_id", "comment2"]

    column_names = AthenaSource(s3_staging_dir=s3_staging_dir,).get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert column_names == expected


@patch("boto3.client")
def test_athena_source_get_table_names_list(mock_connection):
    database_name = "test_database"
    table_name = "test_table"
    s3_staging_dir = "s3_test"
    response = {
        "TableMetadataList": [
            {
                "Name": table_name,
            }
        ]
    }
    mock_connection.return_value = mock_connection
    mock_connection.list_table_metadata.return_value = response
    expected = [table_name]

    table_names_list = AthenaSource(
        s3_staging_dir=s3_staging_dir,
    ).get_table_names_list(database_name=database_name)

    assert table_names_list == expected


@patch("boto3.client")
def test_glue_source_get_column_names(mock_connection):
    database_name = "test_database"
    table_name = "test_table"
    response = {
        "Table": {
            "StorageDescriptor": {
                "Columns": [
                    {"Name": "timestamp", "Type": "string", "Comment": "comment1"},
                    {"Name": "address_id", "Type": "int", "Comment": "comment2"},
                ]
            }
        }
    }
    mock_connection.return_value = mock_connection
    mock_connection.get_table.return_value = response
    expected = ["timestamp", "comment1", "address_id", "comment2"]

    column_names = GlueSource().get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert column_names == expected


@patch("boto3.client")
def test_glue_source_get_table_names_list(mock_connection):
    database_name = "test_database"
    table_name = "test_table"
    response = {
        "TableList": [
            {
                "DatabaseName": database_name,
                "Name": table_name,
            }
        ]
    }
    mock_connection.return_value = mock_connection
    mock_connection.get_tables.return_value = response
    expected = [table_name]

    table_names_list = GlueSource().get_table_names_list(database_name=database_name)

    assert table_names_list == expected



from types import SimpleNamespace
from unittest.mock import Mock, patch

from metadata_guardian.source.external.gcp_source import BigQuerySource


@patch("google.cloud.bigquery.Client.from_service_account_json")
def test_big_query_source_get_column_names(mock_connection):
    service_account_json_path = ""
    dataset_name = "test_dataset"
    table_name = "test_table"
    results = [
        SimpleNamespace(column_name="timestamp", description="description1"),
        SimpleNamespace(column_name="address_id", description="description2"),
    ]
    mock_connection.return_value = mock_connection
    response = Mock()
    response.result.return_value = results
    mock_connection.query.return_value = response
    expected = ["timestamp", "description1", "address_id", "description2"]

    column_names = BigQuerySource(
        service_account_json_path=service_account_json_path
    ).get_column_names(
        database_name=dataset_name, table_name=table_name, include_comment=True
    )

    assert column_names == expected


@patch("google.cloud.bigquery.Client.from_service_account_json")
def test_big_query_source_get_table_names_list(mock_connection):
    service_account_json_path = ""
    dataset_name = "test_dataset"
    results = [
        SimpleNamespace(table_name="test_table"),
        SimpleNamespace(table_name="test_table2"),
    ]
    mock_connection.return_value = mock_connection
    response = Mock()
    response.result.return_value = results
    mock_connection.query.return_value = response
    expected = ["test_table", "test_table2"]

    table_names_list = BigQuerySource(
        service_account_json_path=service_account_json_path
    ).get_table_names_list(
        database_name=dataset_name,
    )

    assert table_names_list == expected



from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional

import snowflake.connector
from loguru import logger
from snowflake.connector import SnowflakeConnection

from .external_metadata_source import ExternalMetadataSource


class SnowflakeAuthenticator(Enum):
    """Authentication method for Snowflake source."""

    USER_PWD = 1
    OKTA = 2
    TOKEN = 3


@dataclass
class SnowflakeSource(ExternalMetadataSource):
    """Instance of a Snowflake source."""

    sf_account: str
    sf_user: str
    sf_password: str
    warehouse: str
    schema_name: str
    okta_account_name: Optional[str] = None
    oauth_token: Optional[str] = None
    oauth_host: Optional[str] = None
    authenticator: SnowflakeAuthenticator = SnowflakeAuthenticator.USER_PWD

    def get_connection(self) -> SnowflakeConnection:
        if self.authenticator == SnowflakeAuthenticator.USER_PWD:
            return snowflake.connector.connect(
                account=self.sf_account,
                user=self.sf_user,
                password=self.sf_password,
                warehouse=self.warehouse,
            )
        elif self.authenticator == SnowflakeAuthenticator.OKTA:
            return snowflake.connector.connect(
                account=self.sf_account,
                user=self.sf_user,
                password=self.sf_password,
                warehouse=self.warehouse,
                authenticator=f"https://{self.okta_account_name}.okta.com/",
            )
        elif self.authenticator == SnowflakeAuthenticator.TOKEN:
            return snowflake.connector.connect(
                account=self.sf_account,
                user=self.sf_user,
                password=self.sf_password,
                warehouse=self.warehouse,
                authenticator="oauth",
                token=self.oauth_token,
            )

    def get_column_names(
        self, database_name: str, table_name: str, include_comment: bool = False
    ) -> List[str]:
        """
        Get column names from the table.
        :param database_name: the database name
        :param table_name: the table name
        :param include_comment: include the comment
        :return: the list of the column names
        """
        try:
            connection = self.get_connection()
            cursor = connection.cursor()
            cursor.execute(
                f'SHOW COLUMNS IN "{database_name}"."{self.schema_name}"."{table_name}"'
            )
            rows = cursor.fetchall()
            columns = list()
            for row in rows:
                column_name = row[2]
                columns.append(column_name.lower())
                if include_comment:
                    column_comment = row[8]
                    columns.append(column_comment.lower())
            return columns
        except Exception as exception:
            logger.exception(
                f"Error in getting columns name from Snowflake {self.schema_name}.{database_name}.{table_name}"
            )
            raise exception
        finally:
            cursor.close()
            connection.close()

    def get_table_names_list(self, database_name: str) -> List[str]:
        """
        Get the table names list from the Snowflake database.
        :param database_name: the database name
        :return: the list of the table names of the database
        """
        try:
            connection = self.get_connection()
            cursor = connection.cursor()
            cursor.execute(f'SHOW TABLES IN DATABASE "{database_name.upper()}"')
            rows = cursor.fetchall()
            table_list = list()
            for row in rows:
                table_name = row[1]
                table_list.append(table_name.upper())
            return table_list
        except Exception as exception:
            logger.exception(
                f"Error in getting table names from the database {database_name} in Snowflake {database_name}"
            )
            raise exception
        finally:
            cursor.close()
            connection.close()

    @property
    def type(self) -> str:
        """
        The type of the source.
        :return: the name of the source.
        """
        return "Snowflake"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import toml

sys.path.insert(0, os.path.abspath("../metadata_guardian/"))
sys.path.insert(0, os.path.abspath("./_ext"))


def get_release_version() -> str:
    """
    Get the release version from the Cargo.toml file

    :return:
    """
    cargo_content = toml.load("../../Cargo.toml")
    return cargo_content["package"]["version"]


# -- Project information -----------------------------------------------------

project = "metadata-guardian"
copyright = "2021 Metadata Guardian contributors"
author = "Florian Valeye"
version = get_release_version()

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "edit_on_github"]
autodoc_typehints = "description"
nitpicky = True
nitpick_ignore = [
    ("py:class", "RawDataRules"),
    ("py:class", "pyarrow._fs.FileSystem"),
    ("py:class", "pyarrow._dataset.Dataset"),
    ("py:class", "pyarrow.lib.Schema"),
    ("py:class", "avro.datafile.DataFileReader"),
    ("py:class", "mypy_boto3_athena.client.AthenaClient"),
    ("py:class", "mypy_boto3_glue.client.GlueClient"),
    ("py:class", "snowflake.connector.connection.SnowflakeConnection"),
    ("py:class", "deltalake.data_catalog.DataCatalog"),
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_logo = "../../../logo.png"
html_favicon = "../../../logo.png"
html_theme_options = {
    "external_links": [],
    "github_url": "https://github.com/fvaleye/MetadataGuardian",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

edit_on_github_project = "fvaleye/MetadataGuardian"
edit_on_github_branch = "main"
page_source_prefix = "python/docs/source"



from unittest.mock import patch

from metadata_guardian.source.external.snowflake_source import (
    SnowflakeAuthenticator,
    SnowflakeSource,
)


@patch("snowflake.connector")
def test_snowflake_source_get_column_names(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    table_name = "test_table"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.description = [["name"], ["phone"]]
    mocked_cursor_one.fetchall.return_value = [
        (database_name, table_name, "column1", "", "", "", "", "", "comment1"),
        (database_name, table_name, "column2", "", "", "", "", "", "comment2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW COLUMNS IN "{database_name}"."{schema_name}"."{table_name}"'
    expected = ["column1", "comment1", "column2", "comment2"]

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )

    column_names = source.get_column_names(
        database_name=database_name, table_name=table_name, include_comment=True
    )

    assert column_names == expected
    assert source.authenticator == SnowflakeAuthenticator.USER_PWD



import asyncio
import os
from unittest.mock import patch

from metadata_guardian.data_rules import AvailableCategory, DataRules
from metadata_guardian.report import MetadataGuardianReport, ReportResults
from metadata_guardian.scanner import ColumnScanner, ContentFilesScanner
from metadata_guardian.source.external.snowflake_source import SnowflakeSource


@patch("snowflake.connector")
def test_column_scanner_table_name(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    table_name = "TEST_TABLE"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.description = [["name"], ["phone"]]
    mocked_cursor_one.fetchall.return_value = [
        (database_name, table_name, "column1", "", "", "", "", "", "comment1"),
        (database_name, table_name, "column2", "", "", "", "", "", "comment2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW COLUMNS IN "{database_name}"."{schema_name}"."{table_name}"'
    expected = MetadataGuardianReport(
        report_results=[
            ReportResults(source=f"{database_name}.{table_name}", results=[])
        ]
    )

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )
    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)

    report = ColumnScanner(data_rules=data_rules).scan_external(
        database_name=database_name, table_name=table_name, source=source
    )

    assert report == expected


@patch("snowflake.connector")
def test_column_scanner_database_name(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    table_name = "TEST_TABLE"
    table_name_2 = "TEST_TABLE_2"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.description = [["name"], ["phone"]]
    mocked_cursor_one.fetchall.return_value = [
        (database_name, table_name, "column1", "", "", "", "", "", "comment1"),
        (database_name, table_name_2, "column2", "", "", "", "", "", "comment2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW COLUMNS IN "{database_name}"."{schema_name}"."{table_name}"'
    expected = MetadataGuardianReport(
        report_results=[
            ReportResults(source=f"{database_name}.{table_name}", results=[]),
            ReportResults(source=f"{database_name}.{table_name_2}", results=[]),
        ]
    )

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )
    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)

    report = ColumnScanner(data_rules=data_rules).scan_external(
        database_name=database_name, source=source
    )

    assert report == expected


@patch("snowflake.connector")
def test_column_scanner_database_name_async(mock_connection):
    database_name = "test_database"
    schema_name = "PUBLIC"
    table_name = "TEST_TABLE"
    table_name_2 = "TEST_TABLE_2"
    sf_account = "sf_account"
    sf_user = "sf_user"
    sf_password = "sf_password"
    warehouse = "warehouse"
    mocked_cursor_one = mock_connection.connect().cursor.return_value
    mocked_cursor_one.description = [["name"], ["phone"]]
    mocked_cursor_one.fetchall.return_value = [
        (database_name, table_name, "column1", "", "", "", "", "", "comment1"),
        (database_name, table_name_2, "column2", "", "", "", "", "", "comment2"),
    ]
    mocked_cursor_one.execute.call_args == f'SHOW COLUMNS IN "{database_name}"."{schema_name}"."{table_name}"'
    expected = MetadataGuardianReport(
        report_results=[
            ReportResults(source=f"{database_name}.{table_name}", results=[]),
            ReportResults(source=f"{database_name}.{table_name_2}", results=[]),
        ]
    )

    source = SnowflakeSource(
        sf_account=sf_account,
        sf_user=sf_user,
        sf_password=sf_password,
        warehouse=warehouse,
        schema_name=schema_name,
    )
    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)

    report = asyncio.run(
        ColumnScanner(data_rules=data_rules).scan_external_async(
            database_name=database_name, source=source
        )
    )

    assert report == expected


def test_local_directory_scan():
    directory_path = os.path.join(os.path.dirname(__file__), "resources")
    file_names_extension = "txt"

    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)

    report = ContentFilesScanner(data_rules=data_rules).scan_directory(
        directory_path=directory_path, file_names_extension=file_names_extension
    )

    assert "resources/inclusion_violation.txt" in str(report)



import pytest

from metadata_guardian.source.local.avro_source import AvroSource


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source(local_file):
    source = AvroSource(local_path=local_file)
    expected = ["name", "favorite_number", "favorite_color"]

    column_names = source.get_column_names()

    assert expected == column_names


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source_namespace(local_file):
    source = AvroSource(local_path=local_file)
    expected = "example.avro"

    assert expected == source.namespace


@pytest.mark.parametrize("local_file", ["users.avro"], indirect=["local_file"])
def test_avro_source_attribute_type(local_file):
    source = AvroSource(local_path=local_file)
    expected = ["string", ["int", "null"], ["string", "null"]]

    field_attribute_type = source.get_field_attribute(attribute_name="type")

    assert expected == field_attribute_type



import pytest

from metadata_guardian.data_rules import AvailableCategory, DataRules
from metadata_guardian.scanner import ColumnScanner, ContentFilesScanner
from metadata_guardian.source.local.avro_schema_source import AvroSchemaSource


@pytest.mark.parametrize(
    "local_file", ["users_avro_schema.json"], indirect=["local_file"]
)
def test_get_data_rules_from_category_pii_with_violation(local_file):
    data_rules = DataRules.from_available_category(category=AvailableCategory.PII)
    source = AvroSchemaSource(local_file)

    md_results = ColumnScanner(data_rules=data_rules).scan_local(source=source)

    assert len(md_results.report_results[0].results) == 1
    assert (
        "tests/resources/users_avro_schema.json" in md_results.report_results[0].source
    )
    result = md_results.report_results[0].results[0]
    assert result.content == "name"
    assert result.category == "PII"
    assert len(result.data_rules) == 1
    data_rule = result.data_rules[0]
    assert data_rule.rule_name == "person"
    assert (
        data_rule.documentation
        == "The person is a personal identifiable information.\n"
    )


@pytest.mark.parametrize(
    "local_file", ["users_avro_schema.json"], indirect=["local_file"]
)
def test_get_data_rules_from_category_inclusion_no_violation(local_file):
    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)
    source = AvroSchemaSource(local_file)

    md_results = ColumnScanner(data_rules=data_rules).scan_local(source=source)

    assert "resources/users_avro_schema.json" in md_results.report_results[0].source
    assert len(md_results.report_results[0].results) == 0


@pytest.mark.parametrize(
    "local_file", ["inclusion_violation.txt"], indirect=["local_file"]
)
def test_get_data_rules_from_category_inclusion_violation_content(local_file):
    data_rules = DataRules.from_available_category(category=AvailableCategory.INCLUSION)

    md_results = ContentFilesScanner(data_rules=data_rules).scan_local_file(local_file)

    assert len(md_results.report_results[0].results) == 1
    assert "resources/inclusion_violation.txt" in md_results.report_results[0].source
    result = md_results.report_results[0].results[0]
    assert (
        result.content
        == "feudal age in that they were not bound to the soil but to the master."
    )
    assert result.category == "INCLUSION"
    assert len(result.data_rules) == 1
    data_rule = result.data_rules[0]
    assert data_rule.rule_name == "master"
    assert (
        data_rule.documentation
        == '"Master–slave" is an offensive and exclusionary metaphor that cannot be detached from American\n'
        "history. Prefer describing a hierarchical relationship between nodes more precisely. Prefer using\n"
        "leader/follower, primary/replica or primary/standby.\n"
    )



from dataclasses import dataclass
from typing import Any, Dict, List

import pyarrow
from pyarrow import Schema
from pyarrow.dataset import Dataset
from pyarrow.fs import FileSystem, LocalFileSystem

from ..metadata_source import MetadataSource


@dataclass
class LocalMetadataSource(MetadataSource):
    """LocalMetadata Source contract."""

    local_path: str
    fs: FileSystem = LocalFileSystem()

    def read(self) -> Dataset:
        """
        Read the source local file.
        :return: the file content
        """
        return pyarrow.dataset.dataset(self.local_path, filesystem=self.fs)

    def schema(self) -> Schema:
        """
        Get the source schema
        :return: the file schema
        """
        return self.read().schema

    def get_column_names(self) -> List[str]:
        """
        Get the column names from the schema.
        :return: the list of the column names
        """
        return [column for column in self.schema().names]

    def type(self) -> str:
        return "LocalFile"
